# ***** BEGIN GPL LICENSE BLOCK *****
#
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation,
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ***** END GPL LICENCE BLOCK *****

import bpy
from mathutils import Vector, Matrix, Quaternion
from itertools import chain

bl_info = {
    "name": "Copy Object to Selected",
    "author" : "Rolando Pera",
    "version" : (0,0,1),
    "blender" : (2, 7, 5),
    "location" : "SpaceBar Search -> Copy Object to Selected",
    "description" :
        "Copy an object an position it on the selected faces or vertices of a target object.",
    "warning" : "Alpha release. Could be very buggy. Proceed with caution.",
    "wiki_url" : "",
    "tracker_url" : "",
    "category": "Object",
}


def copy_object_at_face(target_obj, target_obj_mesh, source_obj, source_obj_mesh, to_mat_world):

    # copy the object at the center of each face with the same orientation as its normal
    for poly in target_obj_mesh.polygons:
        if poly.select == True:
            copy = source_obj.copy()
            scene.objects.link(copy)
            copy_mat_world = copy.matrix_world
    
            bpy.ops.mesh.primitive_circle_add(vertices=3, radius=0.2, fill_type='NGON')
            parent_obj = bpy.context.active_object
            
            for face in source_obj_mesh.polygons:
                if face.select == True:
                    copy_loc = Matrix.Translation(face.center)
                    copy_quat = face.normal.to_track_quat('-Z', 'Y')            
                    copy_mat = copy_mat_world * copy_loc * copy_quat.to_matrix().to_4x4()
                    parent_obj.matrix_world = copy_mat
                    copy.select = True
                    bpy.ops.object.parent_set(type='OBJECT')
                    loc = Matrix.Translation(poly.center)
                    quat = poly.normal.to_track_quat('Z', 'Y')            
                    mat = to_mat_world * loc * quat.to_matrix().to_4x4()
                    parent_obj.matrix_world = mat
                    bpy.ops.object.parent_clear(type='CLEAR_KEEP_TRANSFORM')
                    copy.select = False
                    bpy.ops.object.delete()
    return


def copy_object_at_vertex(target_obj, target_obj_mesh, source_obj, source_obj_mesh, to_mat_world):

    # copy the object at selected vertices with the same orientation as their normals
    for vertex in target_obj_mesh.vertices:
        if vertex.select == True:
            copy = source_obj.copy()
            scene.objects.link(copy)
            copy_mat_world = copy.matrix_world
    
            bpy.ops.mesh.primitive_circle_add(vertices=3, radius=0.2, fill_type='NGON')
            parent_obj = bpy.context.active_object
    
            for face in source_obj_mesh.polygons:
                if face.select == True:
                    copy_loc = Matrix.Translation(face.center)
                    copy_quat = face.normal.to_track_quat('-Z', 'Y')            
                    copy_mat = copy_mat_world * copy_loc * copy_quat.to_matrix().to_4x4()
                    parent_obj.matrix_world = copy_mat
                    copy.select = True
                    bpy.ops.object.parent_set(type='OBJECT')
                    loc = Matrix.Translation(vertex.co)
                    quat = vertex.normal.to_track_quat('Z', 'Y')            
                    mat = to_mat_world * loc * quat.to_matrix().to_4x4()
                    parent_obj.matrix_world = mat
                    bpy.ops.object.parent_clear(type='CLEAR_KEEP_TRANSFORM')
                    copy.select = False
                    bpy.ops.object.delete()
    return


def copy_item_at_face(target_obj, target_obj_mesh, source_obj, to_mat_world):

    # copy the object at the center of each face with the same orientation as its normal
    for poly in target_obj_mesh.polygons:
        if poly.select == True:
            copy = source_obj.copy()
            scene.objects.link(copy)            
            loc = Matrix.Translation(poly.center)
            quat = poly.normal.to_track_quat('-Z', 'Y')            
            mat = to_mat_world * loc * quat.to_matrix().to_4x4()
            copy.matrix_world = mat
    return           


def copy_item_at_vertex(target_obj, target_obj_mesh, source_obj, to_mat_world):

    # copy the object at selected vertices with the same orientation as their normals
    for vertex in target_obj_mesh.vertices:
        if vertex.select == True:
            copy = source_obj.copy()
            scene.objects.link(copy)            
            loc = Matrix.Translation(vertex.co)
            quat = vertex.normal.to_track_quat('-Z', 'Y')            
            mat = to_mat_world * loc * quat.to_matrix().to_4x4()
            copy.matrix_world = mat
    return 
            
                                
class CopyToSelected(bpy.types.Operator):
    """Copy Object to Selected"""
    bl_idname = "object.copy_to_selected"
    bl_label = "Copy Object to Selected"
    bl_options = {'REGISTER', 'UNDO'}
    
    def execute(self, context):        
       
        # some initial values        
        global scene        
        scene = bpy.context.scene
        source_obj_sel = "mesh_obj"
        
        # get the number of objects selected
        sel_objs = bpy.context.selected_objects
        num_sel_objs = len(sel_objs)
        if num_sel_objs != 2:
            self.report({'ERROR'}, "Two objects must be selected (source and target).")
            return {'FINISHED'}
        
        # get the currently selected mesh selection mode
        select_mode = context.tool_settings.mesh_select_mode
        sel_mode_face = select_mode[2]         
                    
        # store the location of 3d cursor
        cursor_location = scene.cursor_location.copy()

        # get current active object, apply scale, set mode to object and deselect it
        target_obj = bpy.context.active_object
        if target_obj.type != 'MESH':
            self.report({'ERROR'}, "The target object is not a Mesh.")
            return {'FINISHED'}
        bpy.ops.object.mode_set(mode='OBJECT')
        bpy.ops.object.make_single_user(type='SELECTED_OBJECTS',  obdata=True)
        bpy.ops.object.transform_apply(location=False, rotation=False, scale=True)
        target_obj_mesh = target_obj.data
        to_mat_world = target_obj.matrix_world        
        bpy.context.active_object.select = False

        sel_obj = bpy.context.selected_objects
        
        # if it is a mesh or a curve, center its origin on selected edit mode element
        for obj in sel_obj:
            scene.objects.active = obj
            source_obj_mesh = obj.data
            if obj.type == 'MESH':                    
                sel_poly_source = [f for f in source_obj_mesh.polygons if f.select]                
                # verify that only one face is selected on source object
                if len(sel_poly_source) != 1:
                    bpy.ops.object.mode_set(mode = 'EDIT')
                    self.report({'ERROR'}, "Select one face (none or more than one selected).")
                    return {'FINISHED'}
                obj_mat_world = obj.matrix_world
                for face in source_obj_mesh.polygons:
                    if face.select:
                        center_loc = obj_mat_world * Vector(face.center)
                        scene.cursor_location = center_loc                                   
                bpy.ops.object.origin_set(type='ORIGIN_CURSOR')                
                
            elif obj.type == 'CURVE':
                sel_points = [p for p in chain(*[s.bezier_points for s in source_obj_mesh.splines])if p.select_control_point]
                bpy.ops.object.mode_set(mode = 'EDIT')
                if len(sel_points) != 1:                    
                    self.report({'ERROR'}, "One control point must be selected (none or more selected).")
                    return {'FINISHED'}   
                bpy.ops.view3d.snap_cursor_to_active()
                bpy.ops.object.mode_set(mode='OBJECT')
                bpy.ops.object.origin_set(type='ORIGIN_CURSOR')
                source_obj_sel = "none_mesh_obj"
            
            else:
                source_obj_sel = "none_mesh_obj"
                            
        # set 3d cursor at origin of world  
        scene.cursor_location = (0.0,0.0,0.0)
        
        # get curently active object to copy
        source_obj = bpy.context.object

        # get what elements are selected on target object and the right function to call
        sel_poly_target = [p for p in target_obj_mesh.polygons if p.select]
        sel_vert_target = [v for v in target_obj_mesh.vertices if v.select]

        if sel_poly_target != [] and source_obj_sel == "mesh_obj":
            if sel_mode_face == True:
                copy_object_at_face(target_obj, target_obj_mesh, source_obj, source_obj_mesh, to_mat_world)
            else:
                copy_object_at_vertex(target_obj, target_obj_mesh, source_obj, source_obj_mesh, to_mat_world)
        elif sel_poly_target != [] and source_obj_sel == "none_mesh_obj":
            if sel_mode_face == True:
                copy_item_at_face(target_obj, target_obj_mesh, source_obj, to_mat_world)
            else:
                copy_item_at_vertex(target_obj, target_obj_mesh, source_obj, to_mat_world)
        elif sel_vert_target != [] and source_obj_sel == "mesh_obj":
            copy_object_at_vertex(target_obj, target_obj_mesh, source_obj, source_obj_mesh, to_mat_world)
        elif sel_vert_target != [] and source_obj_sel == "none_mesh_obj":
            copy_item_at_vertex(target_obj, target_obj_mesh, source_obj, to_mat_world)
        else:
            scene.objects.active = target_obj
            target_obj.select = True
            bpy.ops.object.mode_set(mode = 'EDIT')            
            self.report({'ERROR'}, "At least one face or vertex should be selected.")
            return {'FINISHED'}

        # set 3d cursor back to original location  
        scene.cursor_location = cursor_location
        
        # set source object to active    
        scene.objects.active = source_obj
        
        # set source object into edit mode and selection to faces. Select all the faces
        if source_obj_sel == "mesh_obj":
            bpy.ops.object.mode_set(mode = 'EDIT')
            bpy.context.tool_settings.mesh_select_mode = (False, False, True)
            bpy.ops.mesh.select_all(action='SELECT')
            # change transform to normal and pivot to active element
            for area in bpy.context.screen.areas:
                if area.type == 'VIEW_3D':
                    area.spaces[0].transform_orientation = 'NORMAL'
            bpy.context.space_data.pivot_point = 'ACTIVE_ELEMENT'
        elif source_obj_sel == "none_mesh_obj":
            # change transform to global and pivot to individual origins
            for area in bpy.context.screen.areas:
                if area.type == 'VIEW_3D':
                    area.spaces[0].transform_orientation = 'GLOBAL'
            bpy.context.space_data.pivot_point = 'INDIVIDUAL_ORIGINS'
        
        # end of def execute()    
        return {'FINISHED'}
                    
def register():
    bpy.utils.register_class(CopyToSelected)

def unregister():
    bpy.utils.unregister_class(CopyToSelected)

# to test from text editor without having to install addon.
if __name__ == "__main__":
    register()